@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2014 Junjiro R. Okajima
+ * Copyright (C) 2005-2015 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -25,11 +25,10 @@
 #include <linux/security.h>
 #include "aufs.h"
 
-int au_do_open_nondir(struct file *file, int flags)
+int au_do_open_nondir(struct file *file, int flags, struct file *h_file)
 {
 	int err;
 	aufs_bindex_t bindex;
-	struct file *h_file;
 	struct dentry *dentry;
 	struct au_finfo *finfo;
 	struct inode *h_inode;
@@ -38,11 +37,15 @@ int au_do_open_nondir(struct file *file, int flags)
 
 	err = 0;
 	dentry = file->f_dentry;
+	AuDebugOn(IS_ERR_OR_NULL(dentry));
 	finfo = au_fi(file);
 	memset(&finfo->fi_htop, 0, sizeof(finfo->fi_htop));
 	atomic_set(&finfo->fi_mmapped, 0);
 	bindex = au_dbstart(dentry);
-	h_file = au_h_open(dentry, bindex, flags, file, /*force_wr*/0);
+	if (!h_file)
+		h_file = au_h_open(dentry, bindex, flags, file, /*force_wr*/0);
+	else
+		get_file(h_file);
 	if (IS_ERR(h_file))
 		err = PTR_ERR(h_file);
 	else {
@@ -68,13 +71,16 @@ static int aufs_open_nondir(struct inode *inode __maybe_unused,
 {
 	int err;
 	struct super_block *sb;
+	struct au_do_open_args args = {
+		.open	= au_do_open_nondir
+	};
 
 	AuDbg("%pD, f_flags 0x%x, f_mode 0x%x\n",
 	      file, vfsub_file_flags(file), file->f_mode);
 
 	sb = file->f_dentry->d_sb;
 	si_read_lock(sb, AuLock_FLUSH);
-	err = au_do_open(file, au_do_open_nondir, /*fidir*/NULL);
+	err = au_do_open(file, &args);
 	si_read_unlock(sb);
 	return err;
 }
@@ -122,33 +128,122 @@ static int aufs_flush_nondir(struct file *file, fl_owner_t id)
  * read functions after [fdi]_rwsem are released, but it should be harmless.
  */
 
+/* Callers should call au_read_post() or fput() in the end */
+struct file *au_read_pre(struct file *file, int keep_fi)
+{
+	struct file *h_file;
+	int err;
+
+	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/0);
+	if (!err) {
+		di_read_unlock(file->f_dentry, AuLock_IR);
+		h_file = au_hf_top(file);
+		get_file(h_file);
+		if (!keep_fi)
+			fi_read_unlock(file);
+	} else
+		h_file = ERR_PTR(err);
+
+	return h_file;
+}
+
+static void au_read_post(struct inode *inode, struct file *h_file)
+{
+	/* update without lock, I don't think it a problem */
+	fsstack_copy_attr_atime(inode, file_inode(h_file));
+	fput(h_file);
+}
+
+struct au_write_pre {
+	blkcnt_t blks;
+	aufs_bindex_t bstart;
+};
+
+/*
+ * return with iinfo is write-locked
+ * callers should call au_write_post() or iinfo_write_unlock() + fput() in the
+ * end
+ */
+static struct file *au_write_pre(struct file *file, int do_ready,
+				 struct au_write_pre *wpre)
+{
+	struct file *h_file;
+	struct dentry *dentry;
+	int err;
+	struct au_pin pin;
+
+	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
+	h_file = ERR_PTR(err);
+	if (unlikely(err))
+		goto out;
+
+	dentry = file->f_dentry;
+	if (do_ready) {
+		err = au_ready_to_write(file, -1, &pin);
+		if (unlikely(err)) {
+			h_file = ERR_PTR(err);
+			di_write_unlock(dentry);
+			goto out_fi;
+		}
+	}
+
+	di_downgrade_lock(dentry, /*flags*/0);
+	if (wpre)
+		wpre->bstart = au_fbstart(file);
+	h_file = au_hf_top(file);
+	get_file(h_file);
+	if (wpre)
+		wpre->blks = file_inode(h_file)->i_blocks;
+	if (do_ready)
+		au_unpin(&pin);
+	di_read_unlock(dentry, /*flags*/0);
+
+out_fi:
+	fi_write_unlock(file);
+out:
+	return h_file;
+}
+
+static void au_write_post(struct inode *inode, struct file *h_file,
+			  struct au_write_pre *wpre, ssize_t written)
+{
+	struct inode *h_inode;
+
+	au_cpup_attr_timesizes(inode);
+	AuDebugOn(au_ibstart(inode) != wpre->bstart);
+	h_inode = file_inode(h_file);
+	inode->i_mode = h_inode->i_mode;
+	ii_write_unlock(inode);
+	fput(h_file);
+
+	/* AuDbg("blks %llu, %llu\n", (u64)blks, (u64)h_inode->i_blocks); */
+	if (written > 0)
+		au_fhsm_wrote(inode->i_sb, wpre->bstart,
+			      /*force*/h_inode->i_blocks > wpre->blks);
+}
+
 static ssize_t aufs_read(struct file *file, char __user *buf, size_t count,
 			 loff_t *ppos)
 {
 	ssize_t err;
-	struct dentry *dentry;
+	struct inode *inode;
 	struct file *h_file;
 	struct super_block *sb;
 
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
+	inode = file_inode(file);
+	sb = inode->i_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/0);
-	if (unlikely(err))
-		goto out;
 
-	h_file = au_hf_top(file);
-	get_file(h_file);
-	di_read_unlock(dentry, AuLock_IR);
-	fi_read_unlock(file);
+	h_file = au_read_pre(file, /*keep_fi*/0);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
+		goto out;
 
 	/* filedata may be obsoleted by concurrent copyup, but no problem */
 	err = vfsub_read_u(h_file, buf, count, ppos);
 	/* todo: necessary? */
 	/* file->f_ra = h_file->f_ra; */
-	/* update without lock, I don't think it a problem */
-	fsstack_copy_attr_atime(dentry->d_inode, file_inode(h_file));
-	fput(h_file);
+	au_read_post(inode, h_file);
 
 out:
 	si_read_unlock(sb);
@@ -181,53 +276,24 @@ static ssize_t aufs_write(struct file *file, const char __user *ubuf,
 			  size_t count, loff_t *ppos)
 {
 	ssize_t err;
-	blkcnt_t blks;
-	aufs_bindex_t bstart;
-	struct au_pin pin;
-	struct dentry *dentry;
-	struct inode *inode, *h_inode;
-	struct super_block *sb;
+	struct au_write_pre wpre;
+	struct inode *inode;
 	struct file *h_file;
 	char __user *buf = (char __user *)ubuf;
 
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
-	inode = dentry->d_inode;
+	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
-	if (unlikely(err))
+	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out;
 
-	err = au_ready_to_write(file, -1, &pin);
-	di_downgrade_lock(dentry, AuLock_IR);
-	if (unlikely(err)) {
-		di_read_unlock(dentry, AuLock_IR);
-		fi_write_unlock(file);
-		goto out;
-	}
-
-	bstart = au_fbstart(file);
-	h_file = au_hf_top(file);
-	get_file(h_file);
-	h_inode = h_file->f_dentry->d_inode;
-	blks = h_inode->i_blocks;
-	au_unpin(&pin);
-	di_read_unlock(dentry, AuLock_IR);
-	fi_write_unlock(file);
-
 	err = vfsub_write_u(h_file, buf, count, ppos);
-	ii_write_lock_child(inode);
-	au_cpup_attr_timesizes(inode);
-	inode->i_mode = file_inode(h_file)->i_mode;
-	AuDbg("blks %llu, %llu\n", (u64)blks, (u64)h_inode->i_blocks);
-	if (err > 0)
-		au_fhsm_wrote(sb, bstart, /*force*/h_inode->i_blocks > blks);
-	ii_write_unlock(inode);
-	fput(h_file);
+	au_write_post(inode, h_file, &wpre, err);
 
 out:
-	si_read_unlock(sb);
+	si_read_unlock(inode->i_sb);
 	mutex_unlock(&inode->i_mutex);
 	return err;
 }
@@ -270,28 +336,23 @@ static ssize_t aufs_aio_read(struct kiocb *kio, const struct iovec *iov,
 {
 	ssize_t err;
 	struct file *file, *h_file;
-	struct dentry *dentry;
+	struct inode *inode;
 	struct super_block *sb;
 
 	file = kio->ki_filp;
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
+	inode = file_inode(file);
+	sb = inode->i_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/0);
-	if (unlikely(err))
-		goto out;
 
-	h_file = au_hf_top(file);
-	get_file(h_file);
-	di_read_unlock(dentry, AuLock_IR);
-	fi_read_unlock(file);
+	h_file = au_read_pre(file, /*keep_fi*/0);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
+		goto out;
 
 	err = au_do_aio(h_file, MAY_READ, kio, iov, nv, pos);
 	/* todo: necessary? */
 	/* file->f_ra = h_file->f_ra; */
-	/* update without lock, I don't think it a problem */
-	fsstack_copy_attr_atime(dentry->d_inode, file_inode(h_file));
-	fput(h_file);
+	au_read_post(inode, h_file);
 
 out:
 	si_read_unlock(sb);
@@ -302,53 +363,24 @@ static ssize_t aufs_aio_write(struct kiocb *kio, const struct iovec *iov,
 			      unsigned long nv, loff_t pos)
 {
 	ssize_t err;
-	blkcnt_t blks;
-	aufs_bindex_t bstart;
-	struct au_pin pin;
-	struct dentry *dentry;
-	struct inode *inode, *h_inode;
+	struct au_write_pre wpre;
+	struct inode *inode;
 	struct file *file, *h_file;
-	struct super_block *sb;
 
 	file = kio->ki_filp;
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
-	inode = dentry->d_inode;
+	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
-	if (unlikely(err))
+	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out;
 
-	err = au_ready_to_write(file, -1, &pin);
-	di_downgrade_lock(dentry, AuLock_IR);
-	if (unlikely(err)) {
-		di_read_unlock(dentry, AuLock_IR);
-		fi_write_unlock(file);
-		goto out;
-	}
-
-	bstart = au_fbstart(file);
-	h_file = au_hf_top(file);
-	get_file(h_file);
-	h_inode = h_file->f_dentry->d_inode;
-	blks = h_inode->i_blocks;
-	au_unpin(&pin);
-	di_read_unlock(dentry, AuLock_IR);
-	fi_write_unlock(file);
-
 	err = au_do_aio(h_file, MAY_WRITE, kio, iov, nv, pos);
-	ii_write_lock_child(inode);
-	au_cpup_attr_timesizes(inode);
-	inode->i_mode = file_inode(h_file)->i_mode;
-	AuDbg("blks %llu, %llu\n", (u64)blks, (u64)h_inode->i_blocks);
-	if (err > 0)
-		au_fhsm_wrote(sb, bstart, /*force*/h_inode->i_blocks > blks);
-	ii_write_unlock(inode);
-	fput(h_file);
+	au_write_post(inode, h_file, &wpre, err);
 
 out:
-	si_read_unlock(sb);
+	si_read_unlock(inode->i_sb);
 	mutex_unlock(&inode->i_mutex);
 	return err;
 }
@@ -359,19 +391,18 @@ static ssize_t aufs_splice_read(struct file *file, loff_t *ppos,
 {
 	ssize_t err;
 	struct file *h_file;
-	struct dentry *dentry;
+	struct inode *inode;
 	struct super_block *sb;
 
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
+	inode = file_inode(file);
+	sb = inode->i_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/0);
-	if (unlikely(err))
+
+	h_file = au_read_pre(file, /*keep_fi*/1);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out;
 
-	err = -EINVAL;
-	h_file = au_hf_top(file);
-	get_file(h_file);
 	if (au_test_loopback_kthread()) {
 		au_warn_loopback(h_file->f_dentry->d_sb);
 		if (file->f_mapping != h_file->f_mapping) {
@@ -379,15 +410,12 @@ static ssize_t aufs_splice_read(struct file *file, loff_t *ppos,
 			smp_mb(); /* unnecessary? */
 		}
 	}
-	di_read_unlock(dentry, AuLock_IR);
 	fi_read_unlock(file);
 
 	err = vfsub_splice_to(h_file, ppos, pipe, len, flags);
 	/* todo: necessasry? */
 	/* file->f_ra = h_file->f_ra; */
-	/* update without lock, I don't think it a problem */
-	fsstack_copy_attr_atime(dentry->d_inode, file_inode(h_file));
-	fput(h_file);
+	au_read_post(inode, h_file);
 
 out:
 	si_read_unlock(sb);
@@ -399,52 +427,23 @@ aufs_splice_write(struct pipe_inode_info *pipe, struct file *file, loff_t *ppos,
 		  size_t len, unsigned int flags)
 {
 	ssize_t err;
-	blkcnt_t blks;
-	aufs_bindex_t bstart;
-	struct au_pin pin;
-	struct dentry *dentry;
-	struct inode *inode, *h_inode;
-	struct super_block *sb;
+	struct au_write_pre wpre;
+	struct inode *inode;
 	struct file *h_file;
 
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
-	inode = dentry->d_inode;
+	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
-	if (unlikely(err))
+	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out;
 
-	err = au_ready_to_write(file, -1, &pin);
-	di_downgrade_lock(dentry, AuLock_IR);
-	if (unlikely(err)) {
-		di_read_unlock(dentry, AuLock_IR);
-		fi_write_unlock(file);
-		goto out;
-	}
-
-	bstart = au_fbstart(file);
-	h_file = au_hf_top(file);
-	get_file(h_file);
-	h_inode = h_file->f_dentry->d_inode;
-	blks = h_inode->i_blocks;
-	au_unpin(&pin);
-	di_read_unlock(dentry, AuLock_IR);
-	fi_write_unlock(file);
-
 	err = vfsub_splice_from(pipe, h_file, ppos, len, flags);
-	ii_write_lock_child(inode);
-	au_cpup_attr_timesizes(inode);
-	inode->i_mode = file_inode(h_file)->i_mode;
-	AuDbg("blks %llu, %llu\n", (u64)blks, (u64)h_inode->i_blocks);
-	if (err > 0)
-		au_fhsm_wrote(sb, bstart, /*force*/h_inode->i_blocks > blks);
-	ii_write_unlock(inode);
-	fput(h_file);
+	au_write_post(inode, h_file, &wpre, err);
 
 out:
-	si_read_unlock(sb);
+	si_read_unlock(inode->i_sb);
 	mutex_unlock(&inode->i_mutex);
 	return err;
 }
@@ -453,46 +452,25 @@ static long aufs_fallocate(struct file *file, int mode, loff_t offset,
 			   loff_t len)
 {
 	long err;
-	struct au_pin pin;
-	struct dentry *dentry;
-	struct super_block *sb;
+	struct au_write_pre wpre;
 	struct inode *inode;
 	struct file *h_file;
 
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
-	inode = dentry->d_inode;
+	inode = file_inode(file);
 	au_mtx_and_read_lock(inode);
 
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
-	if (unlikely(err))
+	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out;
 
-	err = au_ready_to_write(file, -1, &pin);
-	di_downgrade_lock(dentry, AuLock_IR);
-	if (unlikely(err)) {
-		di_read_unlock(dentry, AuLock_IR);
-		fi_write_unlock(file);
-		goto out;
-	}
-
-	h_file = au_hf_top(file);
-	get_file(h_file);
-	au_unpin(&pin);
-	di_read_unlock(dentry, AuLock_IR);
-	fi_write_unlock(file);
-
 	lockdep_off();
 	err = do_fallocate(h_file, mode, offset, len);
 	lockdep_on();
-	ii_write_lock_child(inode);
-	au_cpup_attr_timesizes(inode);
-	inode->i_mode = file_inode(h_file)->i_mode;
-	ii_write_unlock(inode);
-	fput(h_file);
+	au_write_post(inode, h_file, &wpre, /*written*/1);
 
 out:
-	si_read_unlock(sb);
+	si_read_unlock(inode->i_sb);
 	mutex_unlock(&inode->i_mutex);
 	return err;
 }
@@ -557,44 +535,27 @@ static unsigned long au_flag_conv(unsigned long flags)
 static int aufs_mmap(struct file *file, struct vm_area_struct *vma)
 {
 	int err;
-	aufs_bindex_t bstart;
 	const unsigned char wlock
 		= (file->f_mode & FMODE_WRITE) && (vma->vm_flags & VM_SHARED);
-	struct dentry *dentry;
 	struct super_block *sb;
 	struct file *h_file;
-	struct au_branch *br;
-	struct au_pin pin;
+	struct inode *inode;
 
 	AuDbgVmRegion(file, vma);
 
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
+	inode = file_inode(file);
+	sb = inode->i_sb;
 	lockdep_off();
 	si_read_lock(sb, AuLock_NOPLMW);
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
-	if (unlikely(err))
+
+	h_file = au_write_pre(file, wlock, /*wpre*/NULL);
+	lockdep_on();
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out;
 
-	if (wlock) {
-		err = au_ready_to_write(file, -1, &pin);
-		di_write_unlock(dentry);
-		if (unlikely(err)) {
-			fi_write_unlock(file);
-			goto out;
-		}
-		au_unpin(&pin);
-	} else
-		di_write_unlock(dentry);
-
-	bstart = au_fbstart(file);
-	br = au_sbr(sb, bstart);
-	h_file = au_hf_top(file);
-	get_file(h_file);
+	err = 0;
 	au_set_mmapped(file);
-	fi_write_unlock(file);
-	lockdep_on();
-
 	au_vm_file_reset(vma, h_file);
 	/*
 	 * we cannot call security_mmap_file() here since it may acquire
@@ -605,21 +566,21 @@ static int aufs_mmap(struct file *file, struct vm_area_struct *vma)
 	 */
 	if (!err)
 		err = h_file->f_op->mmap(h_file, vma);
-	if (unlikely(err))
-		goto out_reset;
-
-	au_vm_prfile_set(vma, file);
-	/* update without lock, I don't think it a problem */
-	fsstack_copy_attr_atime(file_inode(file), file_inode(h_file));
-	goto out_fput; /* success */
-
-out_reset:
+	if (!err) {
+		au_vm_prfile_set(vma, file);
+		fsstack_copy_attr_atime(inode, file_inode(h_file));
+		goto out_fput; /* success */
+	}
 	au_unset_mmapped(file);
 	au_vm_file_reset(vma, file);
+
 out_fput:
-	fput(h_file);
 	lockdep_off();
+	ii_write_unlock(inode);
+	lockdep_on();
+	fput(h_file);
 out:
+	lockdep_off();
 	si_read_unlock(sb);
 	lockdep_on();
 	AuTraceErr(err);
@@ -632,45 +593,29 @@ static int aufs_fsync_nondir(struct file *file, loff_t start, loff_t end,
 			     int datasync)
 {
 	int err;
-	struct au_pin pin;
-	struct dentry *dentry;
+	struct au_write_pre wpre;
 	struct inode *inode;
 	struct file *h_file;
-	struct super_block *sb;
-
-	dentry = file->f_dentry;
-	inode = dentry->d_inode;
-	sb = dentry->d_sb;
-	mutex_lock(&inode->i_mutex);
-	err = si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLM);
-	if (unlikely(err))
-		goto out;
 
 	err = 0; /* -EBADF; */ /* posix? */
 	if (unlikely(!(file->f_mode & FMODE_WRITE)))
-		goto out_si;
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
-	if (unlikely(err))
-		goto out_si;
+		goto out;
 
-	err = au_ready_to_write(file, -1, &pin);
-	di_downgrade_lock(dentry, AuLock_IR);
-	if (unlikely(err))
+	inode = file_inode(file);
+	au_mtx_and_read_lock(inode);
+
+	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out_unlock;
-	au_unpin(&pin);
 
-	err = -EINVAL;
-	h_file = au_hf_top(file);
 	err = vfsub_fsync(h_file, &h_file->f_path, datasync);
-	au_cpup_attr_timesizes(inode);
+	au_write_post(inode, h_file, &wpre, /*written*/0);
 
 out_unlock:
-	di_read_unlock(dentry, AuLock_IR);
-	fi_write_unlock(file);
-out_si:
-	si_read_unlock(sb);
-out:
+	si_read_unlock(inode->i_sb);
 	mutex_unlock(&inode->i_mutex);
+out:
 	return err;
 }
 
@@ -679,28 +624,22 @@ out:
 static int aufs_aio_fsync_nondir(struct kiocb *kio, int datasync)
 {
 	int err;
-	struct au_pin pin;
-	struct dentry *dentry;
+	struct au_write_pre wpre;
 	struct inode *inode;
 	struct file *file, *h_file;
 
-	file = kio->ki_filp;
-	dentry = file->f_dentry;
-	inode = dentry->d_inode;
-	au_mtx_and_read_lock(inode);
-
 	err = 0; /* -EBADF; */ /* posix? */
 	if (unlikely(!(file->f_mode & FMODE_WRITE)))
 		goto out;
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
-	if (unlikely(err))
-		goto out;
 
-	err = au_ready_to_write(file, -1, &pin);
-	di_downgrade_lock(dentry, AuLock_IR);
-	if (unlikely(err))
+	file = kio->ki_filp;
+	inode = file_inode(file);
+	au_mtx_and_read_lock(inode);
+
+	h_file = au_write_pre(file, /*do_ready*/1, &wpre);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out_unlock;
-	au_unpin(&pin);
 
 	err = -ENOSYS;
 	h_file = au_hf_top(file);
@@ -718,16 +657,14 @@ static int aufs_aio_fsync_nondir(struct kiocb *kio, int datasync)
 		if (!err)
 			vfsub_update_h_iattr(&h_file->f_path, /*did*/NULL);
 		/*ignore*/
-		au_cpup_attr_timesizes(inode);
 		mutex_unlock(h_mtx);
 	}
+	au_write_post(inode, h_file, &wpre, /*written*/0);
 
 out_unlock:
-	di_read_unlock(dentry, AuLock_IR);
-	fi_write_unlock(file);
-out:
 	si_read_unlock(inode->sb);
 	mutex_unlock(&inode->i_mutex);
+out:
 	return err;
 }
 #endif
@@ -736,22 +673,19 @@ static int aufs_fasync(int fd, struct file *file, int flag)
 {
 	int err;
 	struct file *h_file;
-	struct dentry *dentry;
 	struct super_block *sb;
 
-	dentry = file->f_dentry;
-	sb = dentry->d_sb;
+	sb = file->f_dentry->d_sb;
 	si_read_lock(sb, AuLock_FLUSH | AuLock_NOPLMW);
-	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/0);
-	if (unlikely(err))
+
+	h_file = au_read_pre(file, /*keep_fi*/0);
+	err = PTR_ERR(h_file);
+	if (IS_ERR(h_file))
 		goto out;
 
-	h_file = au_hf_top(file);
 	if (h_file->f_op->fasync)
 		err = h_file->f_op->fasync(fd, h_file, flag);
-
-	di_read_unlock(dentry, AuLock_IR);
-	fi_read_unlock(file);
+	fput(h_file); /* instead of au_read_post() */
 
 out:
 	si_read_unlock(sb);
@@ -763,7 +697,7 @@ out:
 /* no one supports this operation, currently */
 #if 0
 static ssize_t aufs_sendpage(struct file *file, struct page *page, int offset,
-			     size_t len, loff_t *pos , int more)
+			     size_t len, loff_t *pos, int more)
 {
 }
 #endif
